/*
 * A dataspace server exposing regions of a MMC/SD card.
 *
 * Copyright (C) 2020, 2021, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/cpm-generic.h>
#include <l4/devices/dma-generic.h>
#include <l4/devices/gpio-generic.h>
#include <l4/devices/memory.h>
#include <l4/devices/msc-generic.h>

#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/debugger.h>
#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>

#include <ipc/thread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mem/memory_incremental.h>
#include <fsserver/page_queue_shared.h>
#include <fsserver/pages.h>
#include <resource/resource_server.h>
#include <systypes/env.h>

#include "msc_region_opener.h"
#include "msc_region_operations.h"



/* Common configuration. */

static l4_cap_idx_t icucap;



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else if (start > end)
    return start - index;
  else
    return start;
}



/* Common functions. */

static int init_irq(int num, l4_cap_idx_t irq, l4_uint32_t start, l4_uint32_t end)
{
  /* Create interrupt object. */

  long err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irq));

  if (err)
  {
    printf("Could not create IRQ object: %ld\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(start, end, num),
                             irq));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  return 0;
}



/* Peripheral resources. */

static Cpm_chip *cpm;
static Dma_chip *dma;
static Hw::Gpio_chip *gpio;
static Msc_chip *msc;

/* Obtain an abstraction for the memory card. */

static MscRegionOperations *get_msc_region_operations(const char *machine,
                                                      int msc_channel_num,
                                                      int dma_channel_num,
                                                      int card)
{
  char resource[strlen(machine) + 8];
  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t dma_base = 0, dma_base_end = 0;
  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t msc_base = 0, msc_base_end = 0;
  l4_addr_t msc_phys_base = 0, msc_phys_base_end = 0;
  l4_uint32_t dma_irq_start = 0, dma_irq_end = 0;
  l4_uint32_t msc_irq_start = 0, msc_irq_end = 0;

  icucap = l4re_env_get_cap("icu");

  /* Obtain resource details describing I/O memory. */

  sprintf(resource, "%s-cpm", machine);

  if (get_memory(resource, &cpm_base, &cpm_base_end) < 0)
    return NULL;

  cpm = new_cpm_chip(machine, cpm_base);

  sprintf(resource, "%s-dma", machine);

  if (get_memory(resource, &dma_base, &dma_base_end) < 0)
    return NULL;

  dma = new_dma_chip(machine, dma_base, dma_base_end, cpm);

  if (get_irq(resource, &dma_irq_start, &dma_irq_end) < 0)
    return NULL;

  l4_cap_idx_t dma_irq = l4re_util_cap_alloc();

  if (init_irq(0, dma_irq, dma_irq_start, dma_irq_end))
    return NULL;

  dma->enable();

  sprintf(resource, "%s-gpio", machine);

  if (get_memory(resource, &gpio_base, &gpio_base_end) < 0)
    return NULL;

  sprintf(resource, "%s-msc", machine);

  if (get_memory_complete(resource, &msc_base, &msc_base_end,
                          &msc_phys_base, &msc_phys_base_end) < 0)
    return NULL;

  msc = new_msc_chip(machine, msc_phys_base, msc_base, msc_base_end, cpm);

  if (get_irq(resource, &msc_irq_start, &msc_irq_end) < 0)
    return NULL;

  l4_cap_idx_t msc_irq = l4re_util_cap_alloc();

  if (init_irq(msc_channel_num, msc_irq, msc_irq_start, msc_irq_end))
    return NULL;

  Dma_channel *dma_channel = dma->get_channel(dma_channel_num, dma_irq);
  Msc_channel *msc_channel = msc->get_channel(msc_channel_num, msc_irq, dma_channel);

  /* Configure the memory card peripheral clock.
     NOTE: This needs to be parameterised. */

  enum Clock_identifiers clocks[] = {Clock_msc0, Clock_msc1, Clock_msc2};

  /* Configure the GPIO pins.
     NOTE: This needs to be parameterised and to use a more abstract interface. */

  if (!strcmp(machine, "jz4780"))
  {
    cpm->set_frequency(clocks[msc_channel_num], 50000000);
    gpio = new_gpio_chip(machine, gpio_base, 4);
    gpio->config_pad(20, Hw::Gpio_chip::Function_alt, 0);
    gpio->config_pad(21, Hw::Gpio_chip::Function_alt, 0);
    gpio->config_pad(22, Hw::Gpio_chip::Function_alt, 0);
    gpio->config_pad(23, Hw::Gpio_chip::Function_alt, 0);
    gpio->config_pad(28, Hw::Gpio_chip::Function_alt, 0);
    gpio->config_pad(29, Hw::Gpio_chip::Function_alt, 0);
  }
  else if (!strcmp(machine, "x1600"))
  {
    cpm->set_frequency(clocks[msc_channel_num], 25000000);
    gpio = new_gpio_chip(machine, gpio_base, 2);
    gpio->config_pad(17, Hw::Gpio_chip::Function_alt, 1);
    gpio->config_pad(18, Hw::Gpio_chip::Function_alt, 1);
    gpio->config_pad(19, Hw::Gpio_chip::Function_alt, 1);
    gpio->config_pad(20, Hw::Gpio_chip::Function_alt, 1);
    gpio->config_pad(21, Hw::Gpio_chip::Function_alt, 1);
    gpio->config_pad(22, Hw::Gpio_chip::Function_alt, 1);
  }

  /* Attempt to enable the memory card peripheral. */

  msc_channel->enable();

  if (card >= msc_channel->num_cards())
    return NULL;

  struct dma_region region;

  if (get_dma_region(512, 12, &region))
    return NULL;

  return new MscRegionOperations(msc_channel, card, region);
}



/* Default number of pages for files. */

const unsigned int MEMORY_PAGES = 20;



/* Server program. */

int main(int argc, char *argv[])
{
  l4_debugger_set_object_name(l4re_env()->main_thread, "block_server");
  long err;

  /* Introduce concurrency control. */

  err = ipc_thread_init();

  if (err)
  {
    printf("Initialisation error: %s\n", l4sys_errtostr(err));
    return 1;
  }

  if (argc < 5)
  {
    printf("Need the following:\n\n" \
           "A machine indicator such as jz4780 or x1600.\n" \
           "An MSC channel/peripheral number.\n" \
           "A DMA channel number.\n" \
           "A card number.\n\n" \
           "A number of memory pages can be indicated for the use of the " \
           "server.\n\n" \
           "A named capability from the environment can be specified.\n");
    return 1;
  }

  char *machine = argv[1];
  int msc_channel_num = atoi(argv[2]);
  int dma_channel_num = atoi(argv[3]);
  int card = atoi(argv[4]);

  unsigned int memory_pages = MEMORY_PAGES;

  if (argc > 5)
    memory_pages = atoi(argv[5]);

  const char *server_name = (argc > 6) ? argv[6] : ENV_FILESYSTEM_SERVER_NAME;

  /* Obtain a DMA space for associating allocated memory with physical
     addresses. */

  l4_cap_idx_t dma_space;

  err = get_dma_space(&dma_space);

  if (err)
  {
    printf("Could not obtain DMA space: %s\n", l4sys_errtostr(err));
    return 1;
  }

  MscRegionOperations *ops = get_msc_region_operations(machine,
                                                       msc_channel_num,
                                                       dma_channel_num,
                                                       card);

  if (ops == NULL)
  {
    printf("Could not access memory card peripheral.\n");
    return 1;
  }

  /* Some memory plus infrastructure. */

  MemoryIncremental mem(memory_pages, PAGE_SIZE, dma_space);
  PageQueueShared queue;
  Pages pages(&mem, &queue);
  ResourceRegistry registry(&pages);
  MscRegionOpener opener(&registry, ops);

  /* Register a server associating it with the given object. */

  ResourceServer server(&opener);
  server.bind(server_name);

  printf("Starting server using %d pages...\n", memory_pages);
  server.start();

  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
